"""Application configuration."""
import logging
import os
from distutils.util import strtobool

from dotenv import load_dotenv

log = logging.getLogger(__name__)


def guess_bool(v):
    """Convert given value to boolean value."""
    return strtobool(v) if isinstance(v, str) else bool(v)


load_dotenv()


SECRET_KEY = os.environ.get("SECRET_KEY") or None


SCRIPT_NAME = os.environ.get("SCRIPT_NAME") or ""


DEBUG_PYTHON_HTTP_CLIENT = guess_bool(os.environ.get("DEBUG_PYTHON_HTTP_CLIENT"))

LOG_LEVEL = os.environ.get("LOG_LEVEL", "INFO").upper()
logging.basicConfig(level=LOG_LEVEL)

MATOMO_AUTH_TOKEN = os.getenv("MATOMO_AUTH_TOKEN") or None
MATOMO_BASE_URL = os.getenv("MATOMO_BASE_URL") or None
MATOMO_SITE_ID = os.getenv("MATOMO_SITE_ID") or None
if MATOMO_SITE_ID:
    MATOMO_SITE_ID = int(MATOMO_SITE_ID)



# validata-api -- Validata Web API
# By: Validata Team <pierre.dittgen@jailbreak.paris>
#
# Copyright (C) 2018 OpenDataFrance
# https://git.opendatafrance.net/validata/validata-api
#
# validata-api is free software; you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# validata-api is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


"""Application route handlers."""

import logging
import urllib
from io import BytesIO

import cachetools
import requests
import validata_core
from flasgger import swag_from
from flask import request
from validata_core.helpers import FileContentValidataResource, URLValidataResource

from . import app, config
from .json_errors import abort_json, make_json_response

# Schema cache size (nb of simultaneously stored schemas)
SCHEMA_CACHE_SIZE = 20
# Schema time to live (in seconds)
SCHEMA_CACHE_TTL = 60


log = logging.getLogger(__name__)


def bytes_data(f):
    """Get bytes data from Werkzeug FileStorage instance."""
    iob = BytesIO()
    f.save(iob)
    iob.seek(0)
    return iob.getvalue()


@app.route("/")
def index():
    """Home."""
    apidocs_href = "{}/apidocs".format(config.SCRIPT_NAME)
    apidocs_url = urllib.parse.urljoin(request.url, apidocs_href)
    return make_json_response(
        {
            "apidocs_href": apidocs_href,
            "message": (
                "This is the home page of Validata Web API. "
                f"Its documentation is here: {apidocs_url}"
            ),
        },
        args=None,
    )


specs_dict = {
    "get": {
        "summary": "Validate a tabular file from its URL",
        "parameters": [
            {
                "name": "schema",
                "in": "query",
                "type": "string",
                "format": "url",
                "description": "URL of schema to use for validation",
                "example": "https://git.opendatafrance.net/scdl/deliberations/raw/master/schema.json",
                "required": True,
            },
            {
                "name": "url",
                "in": "query",
                "type": "string",
                "format": "url",
                "description": "URL of tabular file to validate",
                "example": "https://git.opendatafrance.net/scdl/deliberations/raw/v2.0/examples/Deliberations_ok.csv",
                "required": True,
            },
        ],
        "produces": ["application/json"],
    },
    "post": {
        "summary": "Validate an uploaded tabular file",
        "parameters": [
            {
                "name": "schema",
                "in": "formData",
                "type": "string",
                "format": "url",
                "description": "URL of schema to use for validation",
                "example": "https://git.opendatafrance.net/scdl/deliberations/raw/master/schema.json",
                "required": True,
            },
            {
                "name": "file",
                "in": "formData",
                "type": "file",
                "description": "The file to upload",
                "required": True,
            },
        ],
        "consumes": ["multipart/form-data"],
        "produces": ["application/json"],
    },
    "responses": {
        "200": {
            "description": "Validation report",
            "schema": {"type": "object", "required": ["_meta", "schema", "report"]},
            "examples": [
                {
                    "_meta": {"version": "0.2.0"},
                    "schema": {},
                    "report": {
                        "error-count": 1,
                        "preset": "table",
                        "table-count": 1,
                        "tables": [
                            {
                                "error-count": 2,
                                "errors": [
                                    {
                                        "code": "invalid-column-delimiter",
                                        "column-number": None,
                                        "message": None,
                                        "message-data": {
                                            "detected": ";",
                                            "expected": ",",
                                        },
                                        "row-number": None,
                                    },
                                    {
                                        "code": "compare-columns-value",
                                        "column-number": 11,
                                        "message": "La valeur de la colonne PREF_DATE [2017-02-03] devrait être supérieure ou égale à la valeur de la colonne DELIB_DATE [2017-10-15]",
                                        "message-data": {
                                            "column1": "PREF_DATE",
                                            "column2": "DELIB_DATE",
                                            "op": "supérieure ou égale",
                                            "value1": "2017-02-03",
                                            "value2": "2017-10-15",
                                        },
                                        "row-number": 2,
                                    },
                                ],
                                "format": "inline",
                                "headers": [
                                    "COLL_NOM",
                                    "COLL_SIRET",
                                    "DELIB_ID",
                                    "DELIB_DATE",
                                    "DELIB_MATIERE_CODE",
                                    "DELIB_MATIERE_NOM",
                                    "DELIB_OBJET",
                                    "BUDGET_ANNEE",
                                    "BUDGET_NOM",
                                    "PREF_ID",
                                    "PREF_DATE",
                                    "VOTE_EFFECTIF",
                                    "VOTE_REEL",
                                    "VOTE_POUR",
                                    "VOTE_CONTRE",
                                    "VOTE_ABSTENTION",
                                    "DELIB_URL",
                                ],
                                "row-count": 2,
                                "schema": "table-schema",
                                "source": "inline",
                                "time": 0.006,
                                "valid": False,
                            }
                        ],
                        "time": 0.014,
                        "valid": False,
                        "warnings": [],
                    },
                }
            ],
        },
        "400": {
            "description": "Error",
            "schema": {"type": "object", "required": ["_meta", "message"]},
            "examples": [
                {"_meta": {"version": "0.2.0"}, "message": "Unsupported format error"}
            ],
        },
    },
}


@cachetools.cached(cachetools.TTLCache(SCHEMA_CACHE_SIZE, SCHEMA_CACHE_TTL))
def download_schema(schema_url):
    """Download schema by its given url"""
    return requests.get(schema_url).json()


@app.route("/validate", methods={"GET", "POST"})
@swag_from(specs_dict)
def validate():
    """Validate endpoint."""

    if request.method == "GET":
        args = {
            "schema": request.args.get("schema"),
            "url": request.args.get("url"),
        }
    else:
        assert request.method == "POST", request.method
        args = {
            "schema": request.form.get("schema"),
        }

    if not args["schema"]:
        abort_json(400, args, 'Missing or empty "schema" parameter')

    # Download Schema from URL to get control on cache
    # schema json dict is passed to validate function as a dict
    try:
        schema_dict = download_schema(args["schema"])
    except Exception as err:
        abort_json(400, {}, str(err))

    validata_resource = None

    if request.method == "GET":
        # URL validation
        if not args["url"]:
            abort_json(400, args, 'Missing or empty "url" parameter')
        validata_resource = URLValidataResource(args["url"])

    elif request.method == "POST":
        # Uploaded file validation
        f = request.files.get("file")
        if f is None:
            abort_json(400, args, 'Missing or empty "file" parameter')
        validata_resource = FileContentValidataResource(f.filename, bytes_data(f))

    header, rows = validata_resource.extract_tabular_data()
    validation_report = validata_core.validate([header] + rows, schema_dict)

    body = {
        "report": validation_report.to_dict(),
    }

    return make_json_response(body, args)



# validata-api -- Validata Web API
# By: Validata Team <pierre.dittgen@jailbreak.paris>
#
# Copyright (C) 2018 OpenDataFrance
# https://git.opendatafrance.net/validata/validata-api
#
# validata-api is free software; you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# validata-api is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


"""Respond JSON errors instead of HTML. Useful for HTTP/JSON APIs."""

# From https://coderwall.com/p/xq88zg/json-exception-handler-for-flask

import logging

import pkg_resources
import ujson as json
from flask import Response, abort
from toolz import assoc
from werkzeug.exceptions import HTTPException, default_exceptions

log = logging.getLogger(__name__)


def abort_json(status_code, args, message=None):
    if message is None:
        exc = default_exceptions.get(status_code)
        if exc is not None:
            message = exc.description
    response = make_json_response({"message": message}, args, status_code=status_code)
    if status_code == 500:
        log.error((message, args, status_code))
    abort(response)


def error_handler(error):
    status_code = error.code if isinstance(error, HTTPException) else 500
    message = (
        str(error) if isinstance(error, HTTPException) else "Internal server error"
    )
    return make_json_response({"message": message}, args=None, status_code=status_code)


def make_json_response(data, args, status_code=None):
    meta = {
        "validata-api-version": pkg_resources.get_distribution("validata-api").version,
        "validata-core-version": pkg_resources.get_distribution(
            "validata-core"
        ).version,
    }
    if args:
        meta = assoc(meta, "args", args)
    return Response(
        json.dumps(assoc(data, "_meta", meta), sort_keys=True),
        mimetype="application/json",
        status=status_code,
    )


class JsonErrors:
    """
    Respond JSON errors.
    Register error handlers for all HTTP exceptions.

    Special case: when FLASK_DEBUG=1 render HTTP 500 errors as HTML instead of JSON.
    """

    def __init__(self, app=None):
        if app:
            self.init_app(app)

    def init_app(self, app):
        self.app = app
        self.register(HTTPException)
        for code in default_exceptions:
            self.register(code)

    def register(self, exception_or_code, handler=None):
        self.app.errorhandler(exception_or_code)(handler or error_handler)



# validata-api -- Validata Web API
# By: Validata Team <pierre.dittgen@jailbreak.paris>
#
# Copyright (C) 2018 OpenDataFrance
# https://git.opendatafrance.net/validata/validata-api
#
# validata-api is free software; you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# validata-api is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


"""Application definition."""


import http.client
import logging

import pkg_resources
from flasgger import Swagger
from flask import Flask
from flask_cors import CORS

from . import config
from .json_errors import JsonErrors

log = logging.getLogger(__name__)

app = Flask(__name__)
app.secret_key = config.SECRET_KEY

if config.DEBUG_PYTHON_HTTP_CLIENT:
    http.client.HTTPConnection.debuglevel = 1  # type: ignore

# Set Werkzeug's `strict_slashes` option for all routes
# instead of defining it for each route individually.
# From https://stackoverflow.com/a/33285603
app.url_map.strict_slashes = False

JsonErrors(app)

CORS(app)

matomo = None
if config.MATOMO_AUTH_TOKEN and config.MATOMO_BASE_URL and config.MATOMO_SITE_ID:
    from flask_matomo import Matomo

    matomo = Matomo(
        app,
        matomo_url=config.MATOMO_BASE_URL,
        id_site=config.MATOMO_SITE_ID,
        token_auth=config.MATOMO_AUTH_TOKEN,
    )


app.config["SWAGGER"] = {
    "title": "Validata API",
    "description": (
        "This is the documentation of "
        "[Validata Web API](https://git.opendatafrance.net/validata/validata-api). "
        "Each endpoint is listed below, and if you click on it you will see "
        "its parameters. "
        "It's also possible to try these endpoints directly from this page "
        "by setting parameters values in a web form."
    ),
    "basePath": config.SCRIPT_NAME,
    "version": pkg_resources.get_distribution("validata-api").version,
    "uiversion": 3,
    "termsOfService": None,  # TODO Set url once terms-of-service page will be created.
}
Swagger(app)


# Keep this import after app initialisation (to avoid cyclic imports)
from . import route_handlers  # noqa isort:skip



#!/usr/bin/env python3

from pathlib import Path

from setuptools import setup

script_dir = Path(__file__).parent

# Gets the long description from the README.md file
readme_filepath = script_dir / "README.md"
with readme_filepath.open("rt", encoding="utf-8") as fd:
    LONG_DESCRIPTION = fd.read()


setup(
    name="Validata-API",
    version="0.5.0a2",
    description="Validata Web API",
    long_description=LONG_DESCRIPTION,
    long_description_content_type="text/markdown",
    url="https://git.opendatafrance.net/validata/validata-api",
    author="Validata team",
    author_email="admin-validata@jailbreak.paris",
    license="AGPLv3",
    # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        "Development Status :: 5 - Production/Stable",
        # Indicate who your project is intended for
        "Intended Audience :: Developers",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Operating System :: OS Independent",
        # Pick your license as you wish (should match "license" above)
        "License :: OSI Approved :: GNU Affero General Public License v3",
        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        "Programming Language :: Python :: 3",
    ],
    install_requires=[
        "apispec",
        "cachetools",
        "flask",
        "flask-cors",
        "flask-matomo",
        "flasgger",
        "marshmallow",
        "python-dotenv",
        "requests",
        "toml",
        "toolz",
        "ujson",
        "validata_core >= 0.7.0a1",
    ],
    packages=["validata_api"],
    zip_safe=True,
)
